@@ -1,6 +1,7 @@
 package noise
 
 import (
+	"bytes"
 	"encoding/hex"
 	"math"
 	"testing"
@@ -249,6 +250,81 @@ func (NoiseSuite) TestXXRoundtrip(c *C) {
 	c.Assert(string(res), Equals, "worri")
 }
 
+func (NoiseSuite) Test_IXpsk2_Roundtrip(c *C) {
+	cs := NewCipherSuite(DH25519, CipherAESGCM, HashSHA256)
+	rngI := new(RandomInc)
+	rngR := new(RandomInc)
+	*rngR = 1
+
+	staticI, err := cs.GenerateKeypair(rngI)
+	if err != nil {
+		c.Fatal(err)
+	}
+	staticR, err := cs.GenerateKeypair(rngR)
+	if err != nil {
+		c.Fatal(err)
+	}
+
+	psk := []byte("00000000000000000000000000000000")
+
+	hsI, _ := NewHandshakeState(Config{
+		CipherSuite:           cs,
+		Random:                rngI,
+		Pattern:               HandshakeIX,
+		PresharedKeyPlacement: 2,
+		PresharedKey:          psk,
+		Initiator:             true,
+		StaticKeypair:         staticI,
+	})
+	hsR, _ := NewHandshakeState(Config{
+		CipherSuite:           cs,
+		Random:                rngR,
+		Pattern:               HandshakeIX,
+		PresharedKeyPlacement: 2,
+		StaticKeypair:         staticR,
+	})
+
+	// -> e, s
+	msg, _, _, _ := hsI.WriteMessage(nil, nil)
+	c.Assert(msg, HasLen, 96)
+
+	res, _, _, err := hsR.ReadMessage(nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(res, HasLen, 0)
+
+	if !bytes.Equal(hsR.PeerStatic(), staticI.Public) {
+		c.Error("wrong public key from peer")
+	}
+
+	// Look up psk from peer static public key
+
+	// responder should know psk now and set it from the
+	// initiators preshared key
+	if err = hsR.SetPresharedKey(psk); err != nil {
+		c.Fatal(err)
+	}
+	// <- e, dhee, dhse, s, dhes, psk
+	msg, csR0, csR1, _ := hsR.WriteMessage(nil, nil)
+	c.Assert(msg, HasLen, 96)
+	res, csI0, csI1, err := hsI.ReadMessage(nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(res, HasLen, 0)
+
+	// transport I -> R
+	msg, err = csI0.Encrypt(nil, nil, []byte("foo"))
+	c.Assert(err, IsNil)
+	res, err = csR0.Decrypt(nil, nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(string(res), Equals, "foo")
+
+	// transport R -> I
+	msg, err = csR1.Encrypt(nil, nil, []byte("bar"))
+	c.Assert(err, IsNil)
+	res, err = csI1.Decrypt(nil, nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(string(res), Equals, "bar")
+}
+
 func (NoiseSuite) Test_NNpsk0_Roundtrip(c *C) {
 	cs := NewCipherSuite(DH25519, CipherChaChaPoly, HashBLAKE2b)
 	rngI := new(RandomInc)
@@ -518,6 +594,65 @@ func (NoiseSuite) TestHandshakeRollback_rs(c *C) {
 	c.Assert(msg, DeepEquals, expected)
 }
 
+func (NoiseSuite) TestSetNonce(c *C) {
+	cs := NewCipherSuite(DH25519, CipherAESGCM, HashSHA512)
+	rngI := new(RandomInc)
+	rngR := new(RandomInc)
+	*rngR = 1
+
+	hsI, _ := NewHandshakeState(Config{
+		CipherSuite: cs,
+		Random:      rngI,
+		Pattern:     HandshakeNN,
+		Initiator:   true,
+	})
+	hsR, _ := NewHandshakeState(Config{
+		CipherSuite: cs,
+		Random:      rngR,
+		Pattern:     HandshakeNN,
+		Initiator:   false,
+	})
+
+	msg, _, _, _ := hsI.WriteMessage(nil, nil)
+	res, _, _, err := hsR.ReadMessage(nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(res, HasLen, 0)
+
+	msg, csR0, csR1, _ := hsR.WriteMessage(nil, nil)
+	res, csI0, csI1, err := hsI.ReadMessage(nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(res, HasLen, 0)
+
+	c.Assert(csI0.Nonce(), Equals, uint64(0))
+	c.Assert(csI1.Nonce(), Equals, uint64(0))
+	c.Assert(csR0.Nonce(), Equals, uint64(0))
+	c.Assert(csR1.Nonce(), Equals, uint64(0))
+
+	const n = 1234
+	clientMessage := []byte("msg1")
+	csI0.SetNonce(n)
+	msg, err = csI0.Encrypt(nil, nil, clientMessage)
+	c.Assert(err, IsNil)
+	// decrypt with incorrect nonce
+	_, err = csR0.Decrypt(nil, nil, msg)
+	c.Assert(err, NotNil)
+	// decrypt with correct nonce
+	csR0.SetNonce(n)
+	res, err = csR0.Decrypt(nil, nil, msg)
+	c.Assert(err, IsNil)
+	c.Assert(string(clientMessage), Equals, string(res))
+
+	c.Assert(csI0.Nonce(), Equals, uint64(n+1))
+	c.Assert(csI1.Nonce(), Equals, uint64(0))
+	c.Assert(csR0.Nonce(), Equals, uint64(n+1))
+	c.Assert(csR1.Nonce(), Equals, uint64(0))
+
+	serverMessage := []byte("msg2")
+	csR1.SetNonce(MaxNonce + 1)
+	_, err = csR1.Encrypt(nil, nil, serverMessage)
+	c.Assert(err, Equals, ErrMaxNonce)
+}
+
 func (NoiseSuite) TestRekey(c *C) {
 	rng := new(RandomInc)
 
